"""Provides interfaces relating to keyboard focus."""
import asyncio
import enum
import typing

import iterm2.api_pb2
import iterm2.connection
import iterm2.notifications


# pylint: disable=too-few-public-methods
class FocusUpdateApplicationActive:
    """Describes a change in whether the application is active."""
    def __init__(self, active):
        self.__application_active = active

    @property
    def application_active(self) -> bool:
        """:returns: `True` if the application is active or `False` if not."""
        return self.__application_active


class FocusUpdateWindowChanged:
    """Describes a change in which window is focused."""

    class Reason(enum.Enum):
        """Gives the reason for the change"""
        # pylint: disable=line-too-long
        TERMINAL_WINDOW_BECAME_KEY = 0  #: A terminal window received keyboard focus.
        TERMINAL_WINDOW_IS_CURRENT = 1  #: A terminal window is current but some non-terminal window (such as Preferences) has keyboard focus.
        TERMINAL_WINDOW_RESIGNED_KEY = 2  #: A terminal window no longer has keyboard focus.
        # pylint: enable=line-too-long

    def __init__(self, window_id: str, event: Reason):
        self.__window_id = window_id
        self.__event = event

    def __repr__(self):
        return "Window {}: {}".format(
            self.window_id,
            FocusUpdateWindowChanged.Reason(self.event).name)

    @property
    def window_id(self) -> str:
        """:returns: the window ID of the window that changed."""
        return self.__window_id

    @property
    def event(self) -> 'Reason':
        """Describes how the window's focus changed.

        :returns: The reason for the update.
        """
        return self.__event


class FocusUpdateSelectedTabChanged:
    """Describes a change in the selected tab."""
    def __init__(self, tab_id: str):
        self.__tab_id = tab_id

    def __repr__(self):
        return "Tab selected: {}".format(self.tab_id)

    @property
    def tab_id(self) -> str:
        """
        :returns: A tab ID, which is a string.
        """
        return self.__tab_id


class FocusUpdateActiveSessionChanged:
    """Describes a change to the active session within a tab."""
    def __init__(self, session_id: str):
        self.__session_id = session_id

    def __repr__(self):
        return "Session activated: {}".format(self.session_id)

    @property
    def session_id(self) -> str:
        """Returns the active session ID within its tab.

        :returns: A session ID, which is a string.
        """
        return self.__session_id


class FocusUpdate:
    """Describes a change to keyboard focus.

    Up to one of `application_active`, `window_changed`,
    `selected_tab_changed`, or `active_session_changed` will not be `None`."""
    def __init__(
            self,
            application_active: FocusUpdateApplicationActive = None,
            window_changed: FocusUpdateWindowChanged = None,
            selected_tab_changed: FocusUpdateSelectedTabChanged = None,
            active_session_changed: FocusUpdateActiveSessionChanged = None):
        self.__application_active = application_active
        self.__window_changed = window_changed
        self.__selected_tab_changed = selected_tab_changed
        self.__active_session_changed = active_session_changed

    def __repr__(self):
        if self.__application_active:
            return str(self.__application_active)
        if self.__window_changed:
            return str(self.__window_changed)
        if self.__selected_tab_changed:
            return str(self.__selected_tab_changed)
        if self.__active_session_changed:
            return str(self.__active_session_changed)
        return "No Event"

    @property
    def application_active(self) -> typing.Union[
            None, FocusUpdateApplicationActive]:
        """
        :returns: `None` if no change to whether the app is active,
            otherwise :class:`FocusUpdateApplicationActive`"""
        return self.__application_active

    @property
    def window_changed(self) -> typing.Union[None, FocusUpdateWindowChanged]:
        """
        :returns: `None` if no change to the current window, otherwise
            :class:`FocusUpdateWindowChanged`."""
        return self.__window_changed

    @property
    def selected_tab_changed(
            self) -> typing.Union[None, FocusUpdateSelectedTabChanged]:
        """
        :returns: `None` if no change to selected tab, otherwise
            :class:`FocusUpdateSelectedTabChanged`."""
        return self.__selected_tab_changed

    @property
    def active_session_changed(
            self) -> typing.Union[None, FocusUpdateActiveSessionChanged]:
        """
        :returns: `None` if no change to active session, otherwise
            :class:`FocusUpdateActiveSessionChanged`."""
        return self.__active_session_changed


class FocusMonitor:
    """An asyncio context manager for monitoring keyboard focus changes.

    :param connection: A connection to iTerm2.

    .. seealso:: Example ":ref:`mrutabs_example`"
    """
    def __init__(self, connection: iterm2.connection.Connection):
        self.__connection = connection
        self.__queue: typing.List[iterm2.api_pb2.FocusChangedNotification] = []
        self.__future: typing.Optional[asyncio.Future] = None
        self.__token = None

    async def __aenter__(self):
        async def async_callback(
                _connection,
                message: iterm2.api_pb2.FocusChangedNotification):
            """Called when focus changes."""
            self.__queue.append(message)
            future = self.__future
            if future is None:
                # Ignore reentrant calls
                return

            self.__future = None
            if future is not None and not future.done():
                temp = self.__queue[0]
                del self.__queue[0]
                future.set_result(temp)

        self.__token = await (
            iterm2.notifications.async_subscribe_to_focus_change_notification(
                self.__connection,
                async_callback))
        return self

    async def __aexit__(self, exc_type, exc, _tb):
        try:
            await iterm2.notifications.async_unsubscribe(
                self.__connection, self.__token)
        except iterm2.notifications.SubscriptionException:
            pass

    async def async_get_next_update(self) -> FocusUpdate:
        """
        When focus changes, returns an update.

        :returns: A :class:`FocusUpdate` object.

        Example:

        .. code-block:: python

            async with iterm2.FocusMonitor(connection) as monitor:
                while True:
                    update = await monitor.async_get_next_update()
                    if update.selected_tab_changed:
                        print("The active tab is now {}".
                            format(update.selected_tab_changed.tab_id))
        """
        if self.__queue:
            temp = self.__queue[0]
            del self.__queue[0]
            return self.handle_proto(temp)

        future: asyncio.Future = asyncio.Future()
        self.__future = future
        await self.__future
        proto: iterm2.api_pb2.FocusChangedNotification = future.result()
        self.__future = None
        return self.handle_proto(proto)

    # pylint: disable=no-self-use
    def handle_proto(self, proto: iterm2.api_pb2.FocusChangedNotification):
        """Create a FocusUpdate from a protobuf."""
        which = proto.WhichOneof('event')
        if which == 'application_active':
            return FocusUpdate(
                application_active=FocusUpdateApplicationActive(
                    proto.application_active))
        if which == 'window':
            return FocusUpdate(window_changed=FocusUpdateWindowChanged(
                proto.window.window_id,
                FocusUpdateWindowChanged.Reason(proto.window.window_status)))
        if which == 'selected_tab':
            return FocusUpdate(
                selected_tab_changed=FocusUpdateSelectedTabChanged(
                    proto.selected_tab))
        if which == 'session':
            return FocusUpdate(
                active_session_changed=FocusUpdateActiveSessionChanged(
                    proto.session))
        return FocusUpdate()
    # pylint: enable=no-self-use



"""
Provides support for iTerm2 variables, which hold information associated
with various objects such as sessions, tabs, and windows.
"""

import asyncio
import enum
import json
import typing

import iterm2.connection
import iterm2.notifications


class VariableScopes(enum.Enum):
    """Describes the scope in which a variable can be evaluated."""
    SESSION = iterm2.api_pb2.VariableScope.Value("SESSION")  #: Session scope
    TAB = iterm2.api_pb2.VariableScope.Value("TAB")  #: Tab scope
    WINDOW = iterm2.api_pb2.VariableScope.Value("WINDOW")  #: Window scope
    APP = iterm2.api_pb2.VariableScope.Value("APP")  #: Whole-app scope


class VariableMonitor:
    """
    Watches for changes to a variable.

    `VariableMonitor` is a context manager that helps observe changes in iTerm2
    Variables.

   :param connection: The connection to iTerm2.
   :param scope: The scope in which the variable should be evaluated.
   :param name: The variable name.
   :param identifier: A tab, window, or session identifier. Must correspond to
       the passed-in scope. If the scope is `APP` this should be None. If the
       scope is `SESSION` or `WINDOW` the identifier may be "all" or "active".

    .. seealso::
        * Example ":ref:`colorhost_example`"
        * Example ":ref:`theme_example`"

    Example:

      .. code-block:: python

          async with iterm2.VariableMonitor(
                  connection,
                  iterm2.VariableScopes.SESSION,
                  "jobName",
                  my_session.session_id) as mon:
              while True:
                  new_value = await mon.async_get()
                  DoSomething(new_value)
        """
    def __init__(
            self,
            connection: iterm2.connection.Connection,
            scope: VariableScopes,
            name: str,
            identifier: typing.Optional[str]):
        self.__connection = connection
        self.__scope = scope
        self.__name = name
        self.__identifier = identifier
        self.__token = None
        self.__queue: asyncio.Queue = asyncio.Queue(
            loop=asyncio.get_event_loop())

    async def __aenter__(self):
        async def callback(_connection, message):
            """Called when a variable changes."""
            await self.__queue.put(message)

        self.__token = await (
            iterm2.notifications.
            async_subscribe_to_variable_change_notification(
                self.__connection,
                callback,
                self.__scope.value,
                self.__name,
                self.__identifier))
        return self

    async def async_get(self) -> typing.Any:
        """Returns the new value of the variable."""
        result = await self.__queue.get()
        json_new_value = result.json_new_value
        return json.loads(json_new_value)

    async def __aexit__(self, exc_type, exc, _tb):
        try:
            await iterm2.notifications.async_unsubscribe(
                self.__connection, self.__token)
        except iterm2.notifications.SubscriptionException:
            pass



"Abstractions for iTerm2 key bindings."

import enum
import json
import iterm2.keyboard
import iterm2.mainmenu
import iterm2.rpc
import typing

def NoParamConstructor(param):
    return ""


def MenuItemIdentifierConstructor(param):
    lines = param.split("\n")
    title = lines[0]
    if len(lines) > 1:
        identifier = lines[1]
    else:
        identifier = None
    return iterm2.MenuItemIdentifier(title, identifier)

def PasteConfigurationConstructor(param):
    if param is None:
        base64 = False
        wait_for_prompts = False
        tab_transform = PasteConfiguration.TabTransform.NONE
        tab_stop_size = False
        delay = 0.01
        chunk_size = 1024
        convert_newlines = False
        remove_newlines = False
        convert_unicode_punctuation = False
        escape_for_shell = False
        remove_controls = False
        bracket_allowed = True
        use_regex_substitution = False
        regex = ""
        substitution = ""
    else:
        root = json.loads(param)
        base64 = root['Base64']
        wait_for_prompts = root['WaitForPrompts']
        tab_transform = PasteConfiguration.TabTransform(root['TabTransform'])
        tab_stop_size = root['TabStopSize']
        delay = root['Delay']
        chunk_size = root['ChunkSize']
        convert_newlines = root['ConvertNewlines']
        remove_newlines = root['RemoveNewlines']
        convert_unicode_punctuation = root['ConvertUnicodePunctuation']
        escape_for_shell = root['EscapeForShell']
        remove_controls = root['RemoveControls']
        bracket_allowed = root['BracketAllowed']
        use_regex_substitution = root['UseRegexSubstitution']
        regex = root['Regex']
        substitution = root['Substitution']

    return PasteConfiguration(
        base64,
        wait_for_prompts,
        tab_transform,
        tab_stop_size,
        delay,
        chunk_size,
        convert_newlines,
        remove_newlines,
        convert_unicode_punctuation,
        escape_for_shell,
        remove_controls,
        bracket_allowed,
        use_regex_substitution,
        regex,
        substitution)

class PasteConfiguration:
    """Configuration parameters for pasting."""

    class TabTransform(enum.Enum):
        "How to transform tabs."
        NONE = 0
        CONVERT_TO_SPACES = 1
        ESCAPE_WITH_CONTROL_V = 2

    def __init__(
         self,
         base64: bool,
         wait_for_prompts: bool,
         tab_transform: 'TabTransform',
         tab_stop_size: bool,
         delay: float,
         chunk_size: int,
         convert_newlines: bool,
         remove_newlines: bool,
         convert_unicode_punctuation: bool,
         escape_for_shell: bool,
         remove_controls: bool,
         bracket_allowed: bool,
         use_regex_substitution: bool,
         regex: str,
         substitution: str):
        self.__base64 = base64
        self.__wait_for_prompts = wait_for_prompts
        self.__tab_transform = tab_transform
        self.__tab_stop_size = tab_stop_size
        self.__delay = delay
        self.__chunk_size = chunk_size
        self.__convert_newlines = convert_newlines
        self.__remove_newlines = remove_newlines
        self.__convert_unicode_punctuation = convert_unicode_punctuation
        self.__escape_for_shell = escape_for_shell
        self.__remove_controls = remove_controls
        self.__bracket_allowed = bracket_allowed
        self.__use_regex_substitution = use_regex_substitution
        self.__regex = regex
        self.__substitution = substitution


    def _encode(self) -> str:
        param = {
            'Base64': self.__base64,
            'WaitForPrompts': self.__wait_for_prompts,
            'TabTransform': self.__tab_transform.value,
            'TabStopSize': self.__tab_stop_size,
            'Delay': self.__delay,
            'ChunkSize': self.__chunk_size,
            'ConvertNewlines': self.__convert_newlines,
            'RemoveNewlines': self.__remove_newlines,
            'ConvertUnicodePunctuation': self.__convert_unicode_punctuation,
            'EscapeForShell': self.__escape_for_shell,
            'RemoveControls': self.__remove_controls,
            'BracketAllowed': self.__bracket_allowed,
            'UseRegexSubstitution': self.__use_regex_substitution,
            'Regex': self.__regex,
            'Substitution': self.__substitution
        }
        return json.dumps(param)

    @property
    def base64(self) -> bool:
        "Returns whether to base64-encode when pasting."
        return __base64

    @base64.setter
    def base64(self, value: bool):
        "Sets whether to base64-encode when pasting."
        self.__base64 = value

    @property
    def wait_for_prompts(self) -> bool:
        "Returns whether to wait for a shell prompt before pasting."
        return __wait_for_prompts

    @wait_for_prompts.setter
    def wait_for_prompts(self, value: bool):
        "Sets whether to wait for a shell prompt before pasting."
        self.__wait_for_prompts = value

    @property
    def tab_transform(self) -> 'TabTransform':
        "Returns how to convert tabs to strings when pasting."
        return __tab_transform

    @tab_transform.setter
    def tab_transform(self, value: 'TabTransform'):
        "Sets how to convert tabs to strings when pasting."
        self.__tab_transform = value

    @property
    def tab_stop_size(self) -> int:
        "When converting tabs to spaces, this gives the number of spaces per tab."
        return __tab_stop_size

    @tab_stop_size.setter
    def tab_stop_size(self, value: int):
        "When converting tabs to spaces, this gives the number of spaces per tab."
        self.__tab_stop_size = value

    @property
    def delay(self) -> float:
        "How long to wait between chunks (seconds)."
        return __delay

    @delay.setter
    def delay(self, value: float):
        "How long to wait between chunks (seconds)."
        self.__delay = value

    @property
    def chunk_size(self) -> bool:
        "Chunk size to send."
        return __chunk_size

    @chunk_size.setter
    def chunk_size(self, value: bool):
        "Chunk size to send."
        self.__chunk_size = value

    @property
    def convert_newlines(self) -> bool:
        "Convert CRLF and LF to CR?"
        return __convert_newlines

    @convert_newlines.setter
    def convert_newlines(self, value: bool):
        "Convert CRLF and LF to CR?"
        self.__convert_newlines = value

    @property
    def remove_newlines(self) -> bool:
        "Remove all newlines?"
        return __remove_newlines

    @remove_newlines.setter
    def remove_newlines(self, value: bool):
        "Remove all newlines?"
        self.__remove_newlines = value

    @property
    def convert_unicode_punctuation(self) -> bool:
        "Returns whether to convert non-ASCII puncutation to ASCII equivalents when pasting."
        return __convert_unicode_punctuation

    @convert_unicode_punctuation.setter
    def convert_unicode_punctuation(self, value: bool):
        "Sets whether to convert non-ASCII puncutation to ASCII equivalents when pasting."
        self.__convert_unicode_punctuation = value

    @property
    def escape_for_shell(self) -> bool:
        "Returns whether to escape control characters for input to a shell when pasting."
        return __escape_for_shell

    @escape_for_shell.setter
    def escape_for_shell(self, value: bool):
        "Sets whether to escape control characters for input to a shell when pasting."
        self.__escape_for_shell = value

    @property
    def remove_controls(self) -> bool:
        "Returns whether to remove control characters when pasting."
        return __remove_controls

    @remove_controls.setter
    def remove_controls(self, value: bool):
        "Sets whether to remove control characters when pasting."
        self.__remove_controls = value

    @property
    def bracket_allowed(self) -> bool:
        "Returns whether to allow bracketed paste."
        return __bracket_allowed

    @bracket_allowed.setter
    def bracket_allowed(self, value: bool):
        "Sets whether to allow bracketed paste."
        self.__bracket_allowed = value

    @property
    def use_regex_substitution(self) -> bool:
        "Returns whether to perform regular expression substitution. See regex and substitution."
        return __use_regex_substitution

    @use_regex_substitution.setter
    def use_regex_substitution(self, value: bool):
        "Sets whether to perform regular expression substitution. See regex and substitution."
        self.__use_regex_substitution = value

    @property
    def regex(self) -> bool:
        "The regular expression pattern to match. See use_regex_substitution."
        return __regex

    @regex.setter
    def regex(self, value: bool):
        "The regular expression pattern to match. See use_regex_substitution."
        self.__regex = value

    @property
    def substitution(self) -> bool:
        "Replaces matches found by regex. See use_regex_substitution."
        return __substitution

    @substitution.setter
    def substitution(self, value: bool):
        "Replaces matches found by regex. See use_regex_substitution."
        self.__substitution = value


class MoveSelectionUnit(enum.Enum):
    "Units by which the cursor can move when modifying a selection."
    CHAR = 0  #: One cell, or two if a double-width character is present.
    WORD = 1  #: Jump over alphanumerics.
    LINE = 2
    MARK = 3  #: Marks can be set manually or (more often) by shell integration.
    BIG_WORD = 4  #: Like WORD but includes punctuation characters.

    def _encode(self) -> int:
        return self.value


def MoveSelectionUnitConstructor(value: str) -> 'MoveSelectionUnit':
    return MoveSelectionUnit(int(value))

class SnippetIdentifier:
    def __init__(self, value: typing.Union[str, dict]):
        """Creates a SnippetIdentifier.

        :param value: Legacy prefs have a snippet title here. New identifiers
            have a dictionary of {'guid': 'unique identifier'}.
        """
        if isinstance(value, str):
            self.__title = value
            self.__guid = None
        else:
            self.__title = None
            self.__guid = value['guid']

    def _encode(self) -> typing.Union[str,dict]:
        if self.__title is None:
            return { 'guid': self.__guid }
        return self.__title


def parse_binding_param(action: 'BindingAction', param) -> typing.Union[None, iterm2.mainmenu.MenuItemIdentifier, PasteConfiguration, MoveSelectionUnit, SnippetIdentifier]:
    constructor = get_constructor(action)
    return constructor(param)


def get_constructor(action: 'BindingAction'):
    if action == BindingAction.NEXT_SESSION:
        return NoParamConstructor

    if action == BindingAction.NEXT_WINDOW:
        return NoParamConstructor

    if action == BindingAction.PREVIOUS_SESSION:
        return NoParamConstructor

    if action == BindingAction.PREVIOUS_WINDOW:
        return NoParamConstructor

    if action == BindingAction.SCROLL_END:
        return NoParamConstructor

    if action == BindingAction.SCROLL_HOME:
        return NoParamConstructor

    if action == BindingAction.SCROLL_LINE_DOWN:
        return NoParamConstructor

    if action == BindingAction.SCROLL_LINE_UP:
        return NoParamConstructor

    if action == BindingAction.SCROLL_PAGE_DOWN:
        return NoParamConstructor

    if action == BindingAction.SCROLL_PAGE_UP:
        return NoParamConstructor

    if action == BindingAction.ESCAPE_SEQUENCE:
        return str

    if action == BindingAction.HEX_CODE:
        return str

    if action == BindingAction.TEXT:
        return str

    if action == BindingAction.IGNORE:
        return NoParamConstructor

    if action == BindingAction.IR_BACKWARD:
        return NoParamConstructor

    if action == BindingAction.SEND_C_H_BACKSPACE:
        return NoParamConstructor

    if action == BindingAction.SEND_C_QM_BACKSPACE:
        return NoParamConstructor

    if action == BindingAction.SELECT_PANE_LEFT:
        return NoParamConstructor

    if action == BindingAction.SELECT_PANE_RIGHT:
        return NoParamConstructor

    if action == BindingAction.SELECT_PANE_ABOVE:
        return NoParamConstructor

    if action == BindingAction.SELECT_PANE_BELOW:
        return NoParamConstructor

    if action == BindingAction.DO_NOT_REMAP_MODIFIERS:
        return NoParamConstructor

    if action == BindingAction.TOGGLE_FULLSCREEN:
        return NoParamConstructor

    if action == BindingAction.REMAP_LOCALLY:
        return NoParamConstructor

    if action == BindingAction.SELECT_MENU_ITEM:
        return MenuItemIdentifierConstructor

    if action == BindingAction.NEW_WINDOW_WITH_PROFILE:
        return str

    if action == BindingAction.NEW_TAB_WITH_PROFILE:
        return str

    if action == BindingAction.SPLIT_HORIZONTALLY_WITH_PROFILE:
        return str

    if action == BindingAction.SPLIT_VERTICALLY_WITH_PROFILE:
        return str

    if action == BindingAction.NEXT_PANE:
        return NoParamConstructor

    if action == BindingAction.PREVIOUS_PANE:
        return NoParamConstructor

    if action == BindingAction.NEXT_MRU_TAB:
        return NoParamConstructor

    if action == BindingAction.MOVE_TAB_LEFT:
        return NoParamConstructor

    if action == BindingAction.MOVE_TAB_RIGHT:
        return NoParamConstructor

    if action == BindingAction.RUN_COPROCESS:
        return str

    if action == BindingAction.FIND_REGEX:
        return str

    if action == BindingAction.SET_PROFILE:
        return str

    if action == BindingAction.VIM_TEXT:
        return str

    if action == BindingAction.PREVIOUS_MRU_TAB:
        return NoParamConstructor

    if action == BindingAction.LOAD_COLOR_PRESET:
        return str

    if action == BindingAction.PASTE_SPECIAL:
        return PasteConfigurationConstructor

    if action == BindingAction.PASTE_SPECIAL_FROM_SELECTION:
        return PasteConfigurationConstructor

    if action == BindingAction.TOGGLE_HOTKEY_WINDOW_PINNING:
        return NoParamConstructor

    if action == BindingAction.UNDO:
        return NoParamConstructor

    if action == BindingAction.MOVE_END_OF_SELECTION_LEFT:
        return MoveSelectionUnitConstructor

    if action == BindingAction.MOVE_END_OF_SELECTION_RIGHT:
        return MoveSelectionUnitConstructor

    if action == BindingAction.MOVE_START_OF_SELECTION_LEFT:
        return MoveSelectionUnitConstructor

    if action == BindingAction.MOVE_START_OF_SELECTION_RIGHT:
        return MoveSelectionUnitConstructor

    if action == BindingAction.DECREASE_HEIGHT:
        return NoParamConstructor

    if action == BindingAction.INCREASE_HEIGHT:
        return NoParamConstructor

    if action == BindingAction.DECREASE_WIDTH:
        return NoParamConstructor

    if action == BindingAction.INCREASE_WIDTH:
        return NoParamConstructor

    if action == BindingAction.SWAP_PANE_LEFT:
        return NoParamConstructor

    if action == BindingAction.SWAP_PANE_RIGHT:
        return NoParamConstructor

    if action == BindingAction.SWAP_PANE_ABOVE:
        return NoParamConstructor

    if action == BindingAction.SWAP_PANE_BELOW:
        return NoParamConstructor

    if action == BindingAction.FIND_AGAIN_DOWN:
        return NoParamConstructor

    if action == BindingAction.FIND_AGAIN_UP:
        return NoParamConstructor

    if action == BindingAction.TOGGLE_MOUSE_REPORTING:
        return NoParamConstructor

    if action == BindingAction.INVOKE_SCRIPT_FUNCTION:
        return str

    if action == BindingAction.DUPLICATE_TAB:
        return NoParamConstructor

    if action == BindingAction.MOVE_TO_SPLIT_PANE:
        return NoParamConstructor

    if action == BindingAction.SEND_SNIPPET:
        return SnippetIdentifier

    return None

class BindingAction(enum.Enum):
    NEXT_SESSION = 0
    NEXT_WINDOW = 1
    PREVIOUS_SESSION = 2
    PREVIOUS_WINDOW = 3
    SCROLL_END = 4
    SCROLL_HOME = 5
    SCROLL_LINE_DOWN = 6
    SCROLL_LINE_UP = 7
    SCROLL_PAGE_DOWN = 8
    SCROLL_PAGE_UP = 9
    ESCAPE_SEQUENCE = 10
    HEX_CODE = 11
    TEXT = 12
    IGNORE = 13
    IR_BACKWARD = 15
    SEND_C_H_BACKSPACE = 16
    SEND_C_QM_BACKSPACE = 17
    SELECT_PANE_LEFT = 18
    SELECT_PANE_RIGHT = 19
    SELECT_PANE_ABOVE = 20
    SELECT_PANE_BELOW = 21
    DO_NOT_REMAP_MODIFIERS = 22
    TOGGLE_FULLSCREEN = 23
    REMAP_LOCALLY = 24
    SELECT_MENU_ITEM = 25
    NEW_WINDOW_WITH_PROFILE = 26
    NEW_TAB_WITH_PROFILE = 27
    SPLIT_HORIZONTALLY_WITH_PROFILE = 28
    SPLIT_VERTICALLY_WITH_PROFILE = 29
    NEXT_PANE = 30
    PREVIOUS_PANE = 31
    NEXT_MRU_TAB = 32
    MOVE_TAB_LEFT = 33
    MOVE_TAB_RIGHT = 34
    RUN_COPROCESS = 35
    FIND_REGEX = 36
    SET_PROFILE = 37
    VIM_TEXT = 38
    PREVIOUS_MRU_TAB = 39
    LOAD_COLOR_PRESET = 40
    PASTE_SPECIAL = 41
    PASTE_SPECIAL_FROM_SELECTION = 42
    TOGGLE_HOTKEY_WINDOW_PINNING = 43
    UNDO = 44
    MOVE_END_OF_SELECTION_LEFT = 45
    MOVE_END_OF_SELECTION_RIGHT = 46
    MOVE_START_OF_SELECTION_LEFT = 47
    MOVE_START_OF_SELECTION_RIGHT = 48
    DECREASE_HEIGHT = 49
    INCREASE_HEIGHT = 50
    DECREASE_WIDTH = 51
    INCREASE_WIDTH = 52
    SWAP_PANE_LEFT = 53
    SWAP_PANE_RIGHT = 54
    SWAP_PANE_ABOVE = 55
    SWAP_PANE_BELOW = 56
    FIND_AGAIN_DOWN = 57
    FIND_AGAIN_UP = 58
    TOGGLE_MOUSE_REPORTING = 59
    INVOKE_SCRIPT_FUNCTION = 60
    DUPLICATE_TAB = 61
    MOVE_TO_SPLIT_PANE = 62
    SEND_SNIPPET = 63

def decode_key_binding(key: str, obj) -> 'KeyBinding':
    """Convert a key and serialized binding to a :class:`~KeyBinding` object.

    :param key: Binding key, such as the keys in the key_mappings dictionary from a :class:`~Profile`.
    :param obj: A serialized key binding, such as the values in the key_mappings dictionary from a :class:`~Profile`.

    :returns: A :class:`~KeyBinding` object.
    """
    return KeyBinding._make(key, obj)

class KeyBinding:
    """A keyboard shortcut along with an action to perform and any extra
    info that configures the action.

    :param character: The code point of the character associated with the keypress.
    :param modifiers: List of modifier keys that must be pressed to activate the binding
    :param keycode: None or the keycode associated with the key.
    :param action: The action to perform.
    :param param: Action-specific data that configures the behavior.
    :param version: Generally this should be None. For actions that send text, use 1 to get newer escaping semantics (use vim-style escaping instead of only backslash + one of 'n', 'e', 'a', or 't').
    :param label: Label for touch bar actions. Set to None for non-touchbar bindings.
    """
    def __init__(
          self,
          character: int,
          modifiers: [iterm2.keyboard.Modifier],
          keycode: typing.Optional[iterm2.keyboard.Keycode],
          action: 'BindingAction',
          param,
          version: typing.Optional[int],
          label: typing.Optional[str]):
        self.__keycode = keycode
        self.__character = character
        self.__modifiers = 0
        for mod in modifiers:
            self.__modifiers = self.__modifiers | mod.to_cocoa()
        self.__action = action
        self.__parsed_param = param
        if isinstance(param, str):
            self.__encoded_param = param
        else:
            self.__encoded_param = param._encode()
        self.__version = version
        self.__label = label

    def __eq__(self, other):
        return (
             self.__keycode == other.__keycode and
             self.__character == other.__character and
             self.__modifiers == other.__modifiers and
             self.__action == other.__action and
             self.__encoded_param == other.__encoded_param and
             self.__version == other.__version and
             self.__label == other.__label)

    def __repr__(self):
        return f'[KeyBinding keycode={self.keycode} character={self.character} modifiers={self.modifiers} action={self.action} param={self.__parsed_param}] version={self.__version} label={self.__label}]'

    @staticmethod
    def _make(key: str, entry: dict):
        # Key can be one of:
        # 0xcharacter-0xmodifiers
        # 0xcharacter-0xmodifiers-0xkeycode
        keyParts = key.split("-")
        action = BindingAction(entry['Action'])
        if len(keyParts) < 3:
            keycode = None
        else:
            keycode = iterm2.keyboard.Keycode(int(keyParts[2], 16))
        character = int(keyParts[0], 16)
        modifiers = iterm2.keyboard.Modifier.from_cocoa(int(keyParts[1], 16))
        param = parse_binding_param(
                action,
                entry['Text'])
        return KeyBinding(character, modifiers, keycode, action, param, entry.get('Version', None), entry.get('Label', None))

    @property
    def encode(self) -> dict:
        "Encode the key binding so that it can be saved in a profile."
        return self._value

    @property
    def keycode(self) -> typing.Optional[iterm2.keyboard.Keycode]:
        "None, or the keycode associated with the key."
        return self.__keycode

    @property
    def character(self) -> int:
        return self.__character

    @property
    def modifiers(self) -> typing.List['iterm2.keyboard.Modifier']:
        "List of modifier keys that must be pressed to activate the binding."
        return iterm2.keyboard.Modifier.from_cocoa(self.__modifiers)

    @property
    def action(self) -> 'BindingAction':
        "The action to perform."
        return self.__action

    @property
    def param(self) -> typing.Union[None, iterm2.mainmenu.MenuItemIdentifier, PasteConfiguration, MoveSelectionUnit, SnippetIdentifier]:
        "Action-specific data that configures the behavior."
        return self.__parsed_param

    @property
    def key(self):
        """Dictionary key used when setting bindings in a profile."""
        base = hex(self.__character) + "-" + hex(self.__modifiers)
        if self.__keycode is None:
            return base
        return base + "-" + hex(self.__keycode.value)

    @property
    def _value(self) -> dict:
        result = { 'Action': self.action.value,
                   'Text': self.__encoded_param }
        if self.__version is not None:
            result['Version'] = self.__version
        if self.__label is not None:
            result['Label'] = self.__label
        return result

GLOBAL_KEY_MAP_USER_DEFAULTS_KEY = 'GlobalKeyMap'

async def async_get_global_key_bindings(connection: iterm2.connection.Connection) -> [KeyBinding]:
    """Fetches the global key binding.

    :param connection: The :class:`~iterm2.Connection` to use.

    :returns: A list of key bindings from Preferences > Keys.
    """
    proto = await iterm2.rpc.async_get_preference(connection, GLOBAL_KEY_MAP_USER_DEFAULTS_KEY)
    j = proto.preferences_response.results[0].get_preference_result.json_value
    raw = json.loads(j)
    result = []
    for k in raw:
        entry = raw[k]
        binding = decode_key_binding(k, entry)
        result.append(binding)
    return result

async def async_set_global_key_bindings(connection: iterm2.connection.Connection, bindings: [KeyBinding]):
    """Sets the global key bindings.

    :param connection: The :class:`~iterm2.Connection` to use.
    :param bindings: The new bindings to use.
    """
    replacement = {}
    for binding in bindings:
        replacement[binding.key] = binding._value
    proto = await iterm2.rpc.async_set_preference(connection, GLOBAL_KEY_MAP_USER_DEFAULTS_KEY, json.dumps(replacement))
    status = proto.preferences_response.results[0].set_preference_result.status
    if status == iterm2.api_pb2.PreferencesResponse.Result.SetPreferenceResult.Status.Value("OK"):
        return
    raise iterm2.rpc.RPCException(
        iterm2.api_pb2.PreferencesResponse.Result.SetPreferenceResult.Status.Name(status))




"""
Provides classes for monitoring keyboard activity and modifying how iTerm2
handles keystrokes.
"""
import asyncio
import enum
import typing

import iterm2.api_pb2
import iterm2.capabilities
import iterm2.connection
import iterm2.notifications


# pylint: disable=line-too-long
class Modifier(enum.Enum):
    """Enumerated list of modifier keys."""
    CONTROL = iterm2.api_pb2.Modifiers.Value("CONTROL")  #: The control key modifier
    OPTION = iterm2.api_pb2.Modifiers.Value("OPTION")  #: The option (or Alt) key modifier
    COMMAND = iterm2.api_pb2.Modifiers.Value("COMMAND")  #: The command key modifier
    SHIFT = iterm2.api_pb2.Modifiers.Value("SHIFT")  #: The shift key modifier
    FUNCTION = iterm2.api_pb2.Modifiers.Value("FUNCTION")  #: Indicates the key is a function key.
    NUMPAD = iterm2.api_pb2.Modifiers.Value("NUMPAD")  #: Indicates the key is on the numeric keypad.

    @staticmethod
    def from_cocoa(value: int) -> ['Modifier']:
        result = []
        if value & (1 << 18):
            result.append(Modifier.CONTROL)
        if value & (1 << 19):
            result.append(Modifier.OPTION)
        if value & (1 << 20):
            result.append(Modifier.COMMAND)
        if value & (1 << 17):
            result.append(Modifier.SHIFT)
        if value & (1 << 23):
            result.append(Modifier.FUNCTION)
        if value & (1 << 21):
            result.append(Modifier.NUMPAD)
        return result

    def to_cocoa(self) -> int:
        if self == Modifier.CONTROL:
            return 1 << 18

        if self == Modifier.OPTION:
            return 1 << 19

        if self == Modifier.COMMAND:
            return 1 << 20

        if self == Modifier.SHIFT:
            return 1 << 17

        if self == Modifier.FUNCTION:
            return 1 << 23

        if self == Modifier.NUMPAD:
            return 1 << 21

        return 0

# pylint: enable=line-too-long


class Keycode(enum.Enum):
    """Enumerated list of virtual keycodes. These repesent physical keys on a
    keyboard regardless of its actual layout."""
    ANSI_A = 0X00
    ANSI_S = 0X01
    ANSI_D = 0X02
    ANSI_F = 0X03
    ANSI_H = 0X04
    ANSI_G = 0X05
    ANSI_Z = 0X06
    ANSI_X = 0X07
    ANSI_C = 0X08
    ANSI_V = 0X09
    ANSI_B = 0X0B
    ANSI_Q = 0X0C
    ANSI_W = 0X0D
    ANSI_E = 0X0E
    ANSI_R = 0X0F
    ANSI_Y = 0X10
    ANSI_T = 0X11
    ANSI_1 = 0X12
    ANSI_2 = 0X13
    ANSI_3 = 0X14
    ANSI_4 = 0X15
    ANSI_6 = 0X16
    ANSI_5 = 0X17
    ANSI_EQUAL = 0X18
    ANSI_9 = 0X19
    ANSI_7 = 0X1A
    ANSI_MINUS = 0X1B
    ANSI_8 = 0X1C
    ANSI_0 = 0X1D
    ANSI_RIGHT_BRACKET = 0X1E
    ANSI_O = 0X1F
    ANSI_U = 0X20
    ANSI_LEFT_BRACKET = 0X21
    ANSI_I = 0X22
    ANSI_P = 0X23
    ANSI_L = 0X25
    ANSI_J = 0X26
    ANSI_QUOTE = 0X27
    ANSI_K = 0X28
    ANSI_SEMICOLON = 0X29
    ANSI_BACKSLASH = 0X2A
    ANSI_COMMA = 0X2B
    ANSI_SLASH = 0X2C
    ANSI_N = 0X2D
    ANSI_M = 0X2E
    ANSI_PERIOD = 0X2F
    ANSI_GRAVE = 0X32
    ANSI_KEYPAD_DECIMAL = 0X41
    ANSI_KEYPAD_MULTIPLY = 0X43
    ANSI_KEYPAD_PLUS = 0X45
    ANSI_KEYPAD_CLEAR = 0X47
    ANSI_KEYPAD_DIVIDE = 0X4B
    ANSI_KEYPAD_ENTER = 0X4C
    ANSI_KEYPAD_MINUS = 0X4E
    ANSI_KEYPAD_EQUALS = 0X51
    ANSI_KEYPAD0 = 0X52
    ANSI_KEYPAD1 = 0X53
    ANSI_KEYPAD2 = 0X54
    ANSI_KEYPAD3 = 0X55
    ANSI_KEYPAD4 = 0X56
    ANSI_KEYPAD5 = 0X57
    ANSI_KEYPAD6 = 0X58
    ANSI_KEYPAD7 = 0X59
    ANSI_KEYPAD8 = 0X5B
    ANSI_KEYPAD9 = 0X5C
    RETURN = 0X24
    TAB = 0X30
    SPACE = 0X31
    DELETE = 0X33
    ESCAPE = 0X35
    COMMAND = 0X37
    SHIFT = 0X38
    CAPS_LOCK = 0X39
    OPTION = 0X3A
    CONTROL = 0X3B
    RIGHT_COMMAND = 0x36
    RIGHT_SHIFT = 0X3C
    RIGHT_OPTION = 0X3D
    RIGHT_CONTROL = 0X3E
    FUNCTION = 0X3F
    F17 = 0X40
    VOLUME_UP = 0X48
    VOLUME_DOWN = 0X49
    MUTE = 0X4A
    F18 = 0X4F
    F19 = 0X50
    F20 = 0X5A
    F5 = 0X60
    F6 = 0X61
    F7 = 0X62
    F3 = 0X63
    F8 = 0X64
    F9 = 0X65
    F11 = 0X67
    F13 = 0X69
    F16 = 0X6A
    F14 = 0X6B
    F10 = 0X6D
    F12 = 0X6F
    F15 = 0X71
    HELP = 0X72
    HOME = 0X73
    PAGE_UP = 0X74
    FORWARD_DELETE = 0X75
    F4 = 0X76
    END = 0X77
    F2 = 0X78
    PAGE_DOWN = 0X79
    F1 = 0X7A
    LEFT_ARROW = 0X7B
    RIGHT_ARROW = 0X7C
    DOWN_ARROW = 0X7D
    UP_ARROW = 0X7E


class Keystroke:
    """Describes a keystroke.

    Do not create instances of this class. They will be passed to you when you
    use a :class:`KeystrokeMonitor`.
    """

    class Action(enum.Enum):
        """Type of keyboard event."""
        NA = 0  #: Advanced keyboard monitoring is not enabled. Otherwise, this is the same as KEY_DOWN.
        KEY_DOWN = 1  #: A non-modifier was pressed.
        KEY_UP = 2  #: A non-modifier was released.
        FLAGS_CHANGED = 3  #: Only modifiers changed

    def __init__(self, notification):
        self.__characters = notification.characters
        self.__characters_ignoring_modifiers = (
            notification.charactersIgnoringModifiers)
        self.__modifiers = notification.modifiers
        self.__key_code = notification.keyCode
        self.__action = Keystroke.Action.NA
        if notification.HasField("action"):
            if notification.action == iterm2.api_pb2.KeystrokeNotification.Action.Value("KEY_DOWN"):
                self.__action = Keystroke.Action.KEY_DOWN
            elif notification.action == iterm2.api_pb2.KeystrokeNotification.Action.Value("KEY_UP"):
                self.__action = Keystroke.Action.KEY_UP
            elif notification.action == iterm2.api_pb2.KeystrokeNotification.Action.Value("FLAGS_CHANGED"):
                self.__action = Keystroke.Action.FLAGS_CHANGED

    def __repr__(self):
        info = (
            "chars={}, charsIgnoringModifiers={}, " +
            "modifiers={}, keyCode={}").format(
                self.characters,
                self.characters_ignoring_modifiers,
                self.modifiers,
                self.keycode)
        if self.__action == Keystroke.Action.KEY_DOWN:
            info = info + ", action=key-down"
        elif self.__action == Keystroke.Action.KEY_UP:
            info = info + ", action=key-up"
        elif self.__action == Keystroke.Action.FLAGS_CHANGED:
            info = info + ", action=flags-changed"
        return f'Keystroke({info})'

    @property
    def characters(self) -> str:
        """A string giving the characters that would be generated by the
        keystroke ordinarily.

        :returns: A string."""
        return self.__characters

    @property
    def characters_ignoring_modifiers(self) -> str:
        """A string giving the characters that would be generated by the
        keystroke as if no modifiers besides shift were pressed.

        :returns: A string."""
        return self.__characters_ignoring_modifiers

    @property
    def modifiers(self) -> typing.List[Modifier]:
        """The modifiers that were pressed.

        :returns: A list of :class:`Modifier` objects."""
        return list(map(Modifier, self.__modifiers))

    @property
    def keycode(self) -> Keycode:
        """The ANSI keycode that was pressed.

        :returns: A :class:`Keycode` object."""
        return Keycode(self.__key_code)

    @property
    def action(self) -> Action:
        """The kind of keystroke.

        :returns: A :class:`Keystroke.Action` object."""
        return self.__action


class KeystrokePattern:
    """Describes attributes that select keystrokes.

    Keystrokes contain modifiers (e.g., command or option), characters (what
    characters are generated by the keypress), and characters ignoring
    modifiers (what characters would be generated if no modifiers were pressed,
    excepting the shift key).
    """
    def __init__(self):
        self.__required_modifiers = []
        self.__forbidden_modifiers = []
        self.__keycodes = []
        self.__characters = []
        self.__characters_ignoring_modifiers = []

    @property
    def required_modifiers(self) -> typing.List[Modifier]:
        """List of modifiers that are required to match the pattern.

        A list of type :class:`Modifier`.
        """
        return self.__required_modifiers

    @required_modifiers.setter
    def required_modifiers(self, value: typing.List[Modifier]):
        self.__required_modifiers = value

    @property
    def forbidden_modifiers(self) -> typing.List[Modifier]:
        """
        List of modifiers whose presence prevents the pattern from being
        matched.

        A list of type :class:`Modifier`.
        """
        return self.__forbidden_modifiers

    @forbidden_modifiers.setter
    def forbidden_modifiers(self, value: typing.List[Modifier]):
        self.__forbidden_modifiers = value

    @property
    def keycodes(self) -> typing.List[Keycode]:
        """List of keycodes that match the pattern.

        The pattern matches if the modifier constraints are satisfied and a
        keystroke has any of these keycodes.

        A list of type :class:`Keycode`."""
        return self.__keycodes

    @keycodes.setter
    def keycodes(self, value: typing.List[Keycode]):
        self.__keycodes = value

    @property
    def characters(self) -> typing.List[str]:
        """List of strings. Each string has a character.

        The pattern matches if the modifier constraints are satisfied and a
        keystroke has any of these characters.
        """
        return self.__characters

    @characters.setter
    def characters(self, value: typing.List[str]):
        self.__characters = value

    @property
    def characters_ignoring_modifiers(self) -> typing.List[str]:
        """List of strings. Each string has a character.

        The pattern matches if the modifier constraints are satisfied and a
        keystroke has any of these characters, ignoring modifiers.

        "Ignoring modifiers" mostly means ignoring modifiers other than Shift.
        It has a lot of surprising edge cases which Apple did not document, so
        experiment to find how it works.
        """
        return self.__characters_ignoring_modifiers

    @characters_ignoring_modifiers.setter
    def characters_ignoring_modifiers(self, value: typing.List[str]):
        self.__characters_ignoring_modifiers = value

    def to_proto(self):
        """Creates a protobuf for this pattern."""
        # pylint: disable=no-member
        proto = iterm2.api_pb2.KeystrokePattern()
        proto.required_modifiers.extend(
            list(
                map(lambda x: x.value, self.__required_modifiers)))
        proto.forbidden_modifiers.extend(
            list(
                map(lambda x: x.value, self.__forbidden_modifiers)))
        proto.keycodes.extend(list(map(lambda x: x.value, self.__keycodes)))
        proto.characters.extend(self.__characters)
        proto.characters_ignoring_modifiers.extend(
            self.__characters_ignoring_modifiers)
        return proto


class KeystrokeMonitor:
    """Monitors keystrokes in one or all sessions.

    :param connection: The :class:`~iterm2.Connection` to use.
    :param session: The session ID to affect, or `None` meaning all sessions.
    :param advanced: If false only key-down events are reported. If true,
        key-up and flags-changed events are also reported.
    .. seealso::
        * Example ":ref:`broadcast_example`"
        * Example ":ref:`escindicator_example`"
        * Example ":ref:`function_key_tabs_example`"

    Example:

      .. code-block:: python

          async with iterm2.KeystrokeMonitor(connection) as mon:
              while True:
                  keystroke = await mon.async_get()
                  DoSomething(keystroke)
    """
    def __init__(
            self,
            connection: iterm2.connection.Connection,
            session: typing.Union[None, str] = None,
            advanced: typing.Optional[bool] = False):
        self.__connection = connection
        self.__session = session
        if advanced:
            iterm2.capabilities.check_supports_advanced_key_notifications(connection)
        self.__advanced = advanced
        self.__token = None
        self.__queue: asyncio.Queue = asyncio.Queue(
            loop=asyncio.get_event_loop())

    async def __aenter__(self):
        # pylint: disable=unused-argument
        async def callback(connection, notification):
            await self.__queue.put(notification)
        # pylint: enable=unused-argument
        self.__token = (
            await iterm2.notifications.
            async_subscribe_to_keystroke_notification(
                self.__connection,
                callback,
                self.__session,
                self.__advanced))
        return self

    async def async_get(self) -> Keystroke:
        """Wait for and return the next keystroke."""
        notification = await self.__queue.get()
        return Keystroke(notification)

    async def __aexit__(self, exc_type, exc, _tb):
        try:
            await iterm2.notifications.async_unsubscribe(
                self.__connection, self.__token)
        except iterm2.notifications.SubscriptionException:
            pass


class KeystrokeFilter:
    """
    An async context manager that disables the regular handling of keystrokes
    matching patterns during its lifetime.

    :param connection: The :class:`~iterm2.Connection` to use.
    :param patterns: A list of :class:`KeystrokePattern` objects specifying
        keystrokes whose regular handling should be disabled.
    :param session: The session ID to affect, or None meaning all.

    .. seealso::
        * Example ":ref:`broadcast_example`"
        * Example ":ref:`function_key_tabs_example`"

    Example:

    .. code-block:: python

        # Prevent iTerm2 from handling all control-key combinations.
        ctrl = iterm2.KeystrokePattern()
        ctrl.required_modifiers = [iterm2.Modifier.CONTROL]
        ctrl.keycodes = [keycode for keycode in iterm2.Keycode]

        filter = iterm2.KeystrokeFilter(connection, [ctrl])
        async with filter as mon:
            await iterm2.async_wait_forever()
    """
    def __init__(
            self,
            connection: iterm2.connection.Connection,
            patterns: typing.List[KeystrokePattern],
            session: typing.Union[None, str] = None):
        self.__connection = connection
        self.__session = session
        self.__patterns = patterns
        self.__token = None

    async def __aenter__(self):
        self.__token = await iterm2.notifications.async_filter_keystrokes(
            self.__connection,
            self.__patterns,
            self.__session)
        return self

    async def __aexit__(self, exc_type, exc, _tb):
        try:
            await iterm2.notifications.async_unsubscribe(
                self.__connection, self.__token)
        except iterm2.notifications.SubscriptionException:
            pass
